﻿#region License Information
/* HeuristicLab
 * Copyright (C) Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using HEAL.Attic;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Problems.DataAnalysis;

namespace HeuristicLab.Algorithms.DataAnalysis {


  [Item("Spline model (2d)",
    "Bi-variate spline model (wrapper for alglib.spline2sdmodel)")]
  [StorableType("15E4446C-0C15-4B49-8661-51125071CF81")]
  public sealed class Spline2dModel : RegressionModel {
    // not storable! see persistence properties below
    private alglib.spline2dinterpolant interpolant;

    [Storable]
    private string x1;
    [Storable]
    private string x2;
    public override IEnumerable<string> VariablesUsedForPrediction => new[] { x1, x2 };

    [StorableConstructor]
    private Spline2dModel(StorableConstructorFlag deserializing) : base(deserializing) {
      this.interpolant = new alglib.spline2dinterpolant();
    }

    private Spline2dModel(Spline2dModel orig, Cloner cloner) : base(orig, cloner) {
      this.x1 = orig.x1;
      this.x2 = orig.x2;
      this.interpolant = (alglib.spline2dinterpolant)orig.interpolant.make_copy();
    }
    public Spline2dModel(alglib.spline2dinterpolant interpolant, string y, string x1, string x2)
      : base(y, $"Spline model ({x1}, {x2})") {
      this.interpolant = (alglib.spline2dinterpolant)interpolant.make_copy();
      this.x1 = x1;
      this.x2 = x2;
    }


    public override IDeepCloneable Clone(Cloner cloner) => new Spline2dModel(this, cloner);

    public override IRegressionSolution CreateRegressionSolution(IRegressionProblemData problemData) {
      var solution = new RegressionSolution(this, (IRegressionProblemData)problemData.Clone());
      solution.Name = $"Regression Spline ({x1},{x2})";

      return solution;
    }


    public override IEnumerable<double> GetEstimatedValues(IDataset dataset, IEnumerable<int> rows) {
      var x1 = dataset.GetDoubleValues(this.x1, rows);
      var x2 = dataset.GetDoubleValues(this.x2, rows);

      return x1.Zip(x2, (x1i, x2i) => alglib.spline2dcalc(interpolant, x1i, x2i));
    }

    #region persistence
    [Storable]
    public string SerializedInterpolant {
      get {
        alglib.spline2dserialize(interpolant, out var serialized);
        return serialized;
      }
      set {
        alglib.spline2dunserialize(value, out var obj);
        interpolant = obj;
      }
    }
    #endregion
  }
}
